import 'package:flutter/material.dart';
import 'dart:async';
import 'package:adeptutils/src/interfaces/iwidget.dart';
import 'package:adeptutils/src/exception_utils.dart';
import 'package:adeptutils/src/enumerators/lifecycle_state.dart';
import 'package:flutter/scheduler.dart';
import 'package:adeptutils/src/interfaces/ibase.dart';

///This wrapper is used to wrap some useful fields into a final object to be used in the whole StatelessWidget
class StatelessWrapper {
  IBase base;
  BuildContext context;
  LifecycleState lifecycleState;
  Set<StreamSubscription> streamSubscriptions = Set();
}

abstract class AdeptStatelessWidget<T> extends StatelessWidget with WidgetsBindingObserver implements IWidget<T>{
  final _wrapper = StatelessWrapper();

  set _context(BuildContext context) => _wrapper.context = context;
  BuildContext get context => _wrapper.context;

  set _lifecycleState(LifecycleState lifecycleState) => _wrapper.lifecycleState = lifecycleState;
  LifecycleState get lifecycleState => _wrapper.lifecycleState;

  Set<StreamSubscription> get _streamSubscriptions => _wrapper.streamSubscriptions;

  void handleException(dynamic e,
      {
        BuildContext context, String title = "Error",
        String btnContinueText = "CONTINUE",
        String btnExitText = "EXIT"
      }) {
    if(context == null) context = this.context;
    ExceptionUtils.handleException(e, context, title: title,
        btnContinueText: btnContinueText, btnExitText: btnExitText);
  }

  bool addLifecycleAware(StreamSubscription streamSubscription) =>
      streamSubscription != null ? _streamSubscriptions.add(streamSubscription) : false;

  bool removeLifecycleAware(StreamSubscription streamSubscription) =>
      streamSubscription != null ? _streamSubscriptions.remove(streamSubscription) : false;

  void _notifyLifecycleChange(LifecycleState lifecycleState){
    if(lifecycleState == null) return;
    if(lifecycleState != LifecycleState.resumed &&
        lifecycleState != LifecycleState.paused &&
        lifecycleState != LifecycleState.stopped &&
        lifecycleState != LifecycleState.destroyed) return;
    _streamSubscriptions.forEach((streamSubscription) {
      switch(lifecycleState){
        case LifecycleState.resumed:
          streamSubscription.resume();
          break;
        case LifecycleState.paused:
        case LifecycleState.stopped:
          streamSubscription.pause();
          break;
        case LifecycleState.destroyed:
          streamSubscription.cancel();
          break;
        default:
      }
    });
  }

  @mustCallSuper
  @override
  Widget build(BuildContext context) {
    _context = context;
    WidgetsBinding.instance.addObserver(this);
    SchedulerBinding.instance.addPostFrameCallback((_) {
      _notifyLifecycleChange(_lifecycleState = LifecycleState.created);
      onCreated();
      onBuildCompleted();
    });
    return null;
  }

  @override
  void didChangeAppLifecycleState(AppLifecycleState state) {
    if(state == null) return;
    switch(state){
      case AppLifecycleState.resumed:
        _lifecycleState = LifecycleState.resumed;
        _notifyLifecycleChange(lifecycleState);
        onResumed();
        break;
      case AppLifecycleState.inactive:
        _lifecycleState = LifecycleState.inactive;
        _notifyLifecycleChange(lifecycleState);
        onInactive();
        break;
      case AppLifecycleState.paused://AppLifecycleState.paused is used as stopped state because deactivate() works more as a pause for lifecycle
        _lifecycleState = LifecycleState.stopped;
        _notifyLifecycleChange(lifecycleState);
        onStopped();
        break;
      case AppLifecycleState.detached:
        _lifecycleState = LifecycleState.detached;
        _notifyLifecycleChange(lifecycleState);
        onDetached();
        break;
      default:
        break;
    }
  }

  WillPopScope bodyOnBackPressed({
    @required final Widget body,
  }) =>
      WillPopScope(
        child: body,
        onWillPop: onBackPressed,
      );

  @override
  void showLoading(
      {String message, bool showProgressIndeterminate = true, Widget progressIndicator}) {

  }

  @override
  void hideLoading({int delayMillis}) {

  }

  @override
  void extras() {

  }

  @override
  void init() {

  }

  @override
  void check() {

  }

  @override
  void listeners() {

  }

  @override
  bool validate() => true;

  @override
  int getSelectedDrawerItemId() => -1;

  @override
  bool isPressBackAgain() => false;

  @override
  void pressBackAgain(bool pressBackAgain, {String pressBackAgainMessage = "Press back again to exit"}) {
  }

  @override
  void onBuildCompleted() {
  }

  @override
  Future<bool> onBackPressed() => Future(() => true);

  @override
  Future<T> push<T>({BuildContext context, Route<T> route, String routeName, Widget widget}) {
    if(context == null) context = this.context;
    if(route != null) return Navigator.push(context, route); else
    if(widget != null) return Navigator.push(context, MaterialPageRoute(builder: (context) => widget, settings: routeName != null ? RouteSettings(name: routeName) : null));
    return null;
  }

  @override
  Future<T> pushReplacement<T, TO>({BuildContext context, Route<T> newRoute, String newRouteName, Widget newWidget, TO result}) {
    if(context == null) context = this.context;
    if(newRoute != null) return Navigator.pushReplacement(context, newRoute); else
    if(newWidget != null) return Navigator.pushReplacement(context, MaterialPageRoute(builder: (context) => newWidget, settings: newRouteName != null ? RouteSettings(name: newRouteName) : null));
    return null;
  }

  @override
  void onClick({Widget w, id}) {

  }

  @override
  void onLongClick({Widget w, id}) {

  }

  @override
  void onItemLongClick({Widget w, T item, int position, id}) {

  }

  @override
  void onItemClick({Widget w, T item, int position, id}) {

  }

  @override
  void onCreated() {
  }

  @override
  void onResumed() {
  }

  @override
  void onInactive() {
  }

  @override
  void onPaused() {
  }

  @override
  void onDetached() {
  }

  @override
  void onStopped() {
  }

  @override
  void onDestroy() {
  }

}