
typedef OnLocationChanged = Function(Location location);
typedef OnStatusChanged = Function(String provider, int status, dynamic extras);
typedef OnProviderEnabled = Function(String provider);
typedef OnProviderDisabled = Function(String provider);

abstract class LocationListener {


 /// Called when the location has changed.
 /// <p> There are no restrictions on the use of the supplied Location object.
 /// @param location The new location, as a Location object.
  void onLocationChanged(Location location);

  /**
   * Called when the provider status changes. This method is called when
   * a provider is unable to fetch a location or if the provider has recently
   * become available after a period of unavailability.
   *
   * @param provider the name of the location provider associated with this
   * update.
   * @param status {@link LocationProvider#OUT_OF_SERVICE} if the
   * provider is out of service, and this is not expected to change in the
   * near future; {@link LocationProvider#TEMPORARILY_UNAVAILABLE} if
   * the provider is temporarily unavailable but is expected to be available
   * shortly; and {@link LocationProvider#AVAILABLE} if the
   * provider is currently available.
   * @param extras an optional Bundle which will contain provider specific
   * status variables.
   *
   * <p> A number of common key/value pairs for the extras Bundle are listed
   * below. Providers that use any of the keys on this list must
   * provide the corresponding value as described below.
   *
   * <ul>
   * <li> satellites - the number of satellites used to derive the fix
   * </ul>
   */
  void onStatusChanged(String provider, int status, dynamic extras);

  /**
   * Called when the provider is enabled by the user.
   *
   * @param provider the name of the location provider associated with this
   * update.
   */
  void onProviderEnabled(String provider);

  /**
   * Called when the provider is disabled by the user. If requestLocationUpdates
   * is called on an already disabled provider, this method is called
   * immediately.
   *
   * @param provider the name of the location provider associated with this
   * update.
   */
  void onProviderDisabled(String provider);
}

class Location {
  Location({
    this.longitude,
    this.latitude,
    this.timestamp,
    this.mocked,
    this.accuracy,
    this.altitude,
    this.heading,
    this.speed,
    this.speedAccuracy,
  });

  Location._({
    this.longitude,
    this.latitude,
    this.timestamp,
    this.mocked,
    this.accuracy,
    this.altitude,
    this.heading,
    this.speed,
    this.speedAccuracy,
  });

  /// The latitude of this location in degrees normalized to the interval -90.0 to +90.0 (both inclusive).
  final double latitude;

  /// The longitude of the location in degrees normalized to the interval -180 (exclusive) to +180 (inclusive).
  final double longitude;

  /// The time at which this location was determined.
  final DateTime timestamp;

  ///Indicate if location was created from a mock provider.
  ///
  /// The mock information is not available on all devices. In these cases the returned value is false.
  final bool mocked;

  /// The altitude of the device in meters.
  ///
  /// The altitude is not available on all devices. In these cases the returned value is 0.0.
  final double altitude;

  /// The estimated horizontal accuracy of the location in meters.
  ///
  /// The accuracy is not available on all devices. In these cases the value is 0.0.
  final double accuracy;

  /// The heading in which the device is traveling in degrees.
  ///
  /// The heading is not available on all devices. In these cases the value is 0.0.
  final double heading;

  /// The speed at which the devices is traveling in meters per second over ground.
  ///
  /// The speed is not available on all devices. In these cases the value is 0.0.
  final double speed;

  /// The estimated speed accuracy of this location, in meters per second.
  ///
  /// The speedAccuracy is not available on all devices. In these cases the value is 0.0.
  final double speedAccuracy;

  @override
  String toString() {
    return 'Lat: $latitude, Long: $longitude';
  }

  /// Converts a collection of [Map] objects into a collection of [Location] objects.
  static List<Location> _fromMaps(dynamic message) {
    if (message == null) {
      throw ArgumentError('The parameter \'message\' should not be null.');
    }

    final List<Location> list = message.map<Location>(_fromMap).toList();
    return list;
  }

  /// Converts the supplied [Map] to an instance of the [Location] class.
  static Location _fromMap(dynamic message) {
    if (message == null) {
      throw ArgumentError('The parameter \'message\' should not be null.');
    }

    final Map<dynamic, dynamic> locationMap = message;

    if (!locationMap.containsKey('latitude')) {
      throw ArgumentError.value(locationMap, 'locationMap',
          'The supplied map doesn\'t contain the mandatory key `latitude`.');
    }

    if (!locationMap.containsKey('longitude')) {
      throw ArgumentError.value(locationMap, 'locationMap',
          'The supplied map doesn\'t contain the mandatory key `longitude`.');
    }

    final DateTime timestamp = locationMap['timestamp'] != null
        ? DateTime.fromMillisecondsSinceEpoch(locationMap['timestamp'].toInt(),
        isUtc: true)
        : null;

    return Location._(
        latitude: locationMap['latitude'],
        longitude: locationMap['longitude'],
        timestamp: timestamp,
        mocked: locationMap['mocked'] ?? false,
        altitude: locationMap['altitude'] ?? 0.0,
        accuracy: locationMap['accuracy'] ?? 0.0,
        heading: locationMap['heading'] ?? 0.0,
        speed: locationMap['speed'] ?? 0.0,
        speedAccuracy: locationMap['speed_accuracy'] ?? 0.0);
  }
}