import 'package:json_annotation/json_annotation.dart';
import 'package:date_format/date_format.dart' as DateFormatter;
import 'package:adeptutils/src/wrapper/adept_date_time.dart';
import 'package:intl/intl.dart' as Intl;

/// Created by Luis on 03/08/19.

/// This wrapper class is necessary because java.time.ZonedDateTime has issues with
/// Gson serialization, which requires a custom TypeAdapter to serialize/deserialize it...
/// But more important is that ZoneDateTime parsing from String requires Android min API level 26
/// when this project has defined min API level 21.
/// 
/// Default date format example: 2001-07-04T12:08:56.867-07:00
/// Default formatted date format example: Wed, 4 Jul 2001 12:08:56 PM


class ZonedDateTime extends AdeptDateTime
{
  static const String FULL_ZONED_DATE_TIME_FORMAT = "yyyy-MM-dd'T'HH:mm:ss.SSSXXX";
  static const String SIMPLE_ZONED_DATE_TIME_FORMAT = "yyyy-MM-dd'T'HH:mm:ssXXX";

  static const List<String> THREE_LETTER_ISO_8601_TIME_ZONE_FORMAT = ['yyyy','-','mm','-','dd','T','HH',':','nn',':','ss','.','SSS']; ///here the ZoneOffset is managed manually to add the colon(:) between two digit hours and two digit minutes
  static const List<String> TWO_LETTER_ISO_8601_TIME_ZONE_FORMAT = ['yyyy','-','mm','-','dd','T','HH',':','nn',':','ss','.','SSS','z'];
  static const List<String> FORMATTED_DATE_FORMAT_LIST = ['D',', ', 'd', ' ', 'M', ' ' ,'yyyy', ' ','hh',':','nn',':','ss', ' ', "am"];
  static const List<String> FORMATTED_TIME_FORMAT_LIST = ['hh',':','nn',':','ss', ' ', "am"];
  @JsonKey(ignore: true) List<String> dateFormatList = ['yyyy','-','mm','-','dd','T','HH',':','nn',':','ss','.','SSS'];
  @JsonKey(ignore: true) List<String> formattedDateFormatList = FORMATTED_DATE_FORMAT_LIST;
  @JsonKey(ignore: true) List<String> formattedTimeFormatList = FORMATTED_TIME_FORMAT_LIST;

  ZonedDateTime({
    String date,
    int dateTimeMillis,
    DateTime dateAsDateTime,
    bool use24h,
    String dateFormatStandard,
    String formattedDateTimeFormat,
    String formattedDateFormat,
    String formattedTimeFormat,
    List<String> dateFormatList,
    List<String> formattedDateFormatList,
    List<String> formattedTimeFormatList,
  }) : super(
    date: date,
    dateTimeMillis: dateTimeMillis,
    dateAsDateTime: dateAsDateTime,
    use24h: use24h,
    dateFormatStandard: dateFormatStandard,
    formattedDateTimeFormat: formattedDateTimeFormat,
    formattedDateFormat: formattedDateFormat,
    formattedTimeFormat: formattedTimeFormat,
  ) {
    dateFormatList = dateFormatList ?? THREE_LETTER_ISO_8601_TIME_ZONE_FORMAT;
    formattedDateFormatList = formattedDateFormatList ?? FORMATTED_DATE_FORMAT_LIST;
    formattedTimeFormatList = formattedTimeFormatList ?? FORMATTED_TIME_FORMAT_LIST;
    if(date == null){
      if(dateAsDateTime != null) this.dateAsDateTime = dateAsDateTime;
      if(dateTimeMillis != null) this.dateTimeMillis = dateTimeMillis;
    }
  }

  bool get hasZoneOffset {
    try{
      /// ZoneOffset could be as:
      /// <time>Z
      /// <time>±hh:mm
      /// <time>±hhmm
      /// <time>±hh
      return date.toLowerCase().endsWith("z") ||
            date.contains("-", date.length - 6) ||
            date.contains("+", date.length - 6)
      ;
    } catch(e){
      print(e);
    }
    return false;
  }

  String get localZoneOffsetAsString {
    try{
      Duration localZoneOffset = DateTime.now().timeZoneOffset;
      int hours = localZoneOffset.inHours.abs();
      int minutes = localZoneOffset.inMinutes.abs() - (hours * 60);
      String offset = "${localZoneOffset.isNegative ? "-" : "+"}"
          "${Intl.NumberFormat("#00", "en_US").format(hours)}:"
          "${Intl.NumberFormat("#00", "en_US").format(minutes)}";
      return offset;
    } catch(e){
      print(e);
    }
    return "+00:00";
  }

  DateTime _toLocalZone(DateTime dateTime, String date){
    try{
      if(hasZoneOffset) {
        String offset;
        if(date.toLowerCase().endsWith("z"))
          offset = "+00:00";
        else {
          int signIndex = date.lastIndexOf('+');
          if(signIndex < 0) signIndex = date.lastIndexOf('-');
          offset = date.substring(signIndex);
        }
        List split = offset.split(':');
        Duration dateZoneOffset = Duration(hours: int.tryParse(split.length > 0 ? split[0] : "0") ?? 0, minutes: int.tryParse(split.length > 1 ? split[1] : "0") ?? 0);
        Duration localZoneOffset = dateTime.timeZoneOffset;
        int dateZoneOffsetInMinutes = dateZoneOffset.inMinutes;
        int localZoneOffsetInMinutes = localZoneOffset.inMinutes;

        int diffInMinutes = 0;

        if(dateZoneOffsetInMinutes >= localZoneOffsetInMinutes)//This means server date is after the local date
          diffInMinutes = -(dateZoneOffsetInMinutes - localZoneOffsetInMinutes);//Then the difference in minutes should be substracted from local date that's why the minus is in front
        else//This means server date is before the local date
          diffInMinutes = localZoneOffsetInMinutes - dateZoneOffsetInMinutes;//Then the difference in minutes should be added to local date

        return dateTime.add(Duration(minutes: diffInMinutes));
      }
    } catch(e){
      print(e);
    }
    return dateTime;
  }

  @override
  DateTime getDateAsDateTime() {
    if(dateAsDateTime == null){
      if(date != null) {
        try{
          dateAsDateTime = Intl.DateFormat(dateFormat).parse(date);
        }catch(e) {
          print(e);
          dateAsDateTime = DateTime.tryParse(date);
        }
      }
      if(dateAsDateTime == null) setDateAsDateTime(DateTime.now());
      dateAsDateTime = _toLocalZone(dateAsDateTime, date);
    }
    if(date == null) setDateAsDateTime(dateAsDateTime);
    return dateAsDateTime;
  }

  @override
  String getDate() {
    if(date == null) getDateAsDateTime();///this generates a non null date
    if(!hasZoneOffset) date = "$date$localZoneOffsetAsString";
    return date;
  }

  List<String> getDateFormatList() => dateFormatList ?? (dateFormatList = THREE_LETTER_ISO_8601_TIME_ZONE_FORMAT);
  setDateFormatList(List<String> value) => dateFormatList ??= value;

  List<String> getFormattedDateFormatList() => formattedDateFormatList ?? (formattedDateFormatList = FORMATTED_DATE_FORMAT_LIST);
  setFormattedDateFormatList(List<String> value) => formattedDateFormatList = value ?? FORMATTED_DATE_FORMAT_LIST;

  List<String> getFormattedTimeFormatList() => formattedTimeFormatList ?? (formattedTimeFormatList = FORMATTED_TIME_FORMAT_LIST);
  setFormattedTimeFormatList(List<String> value) => formattedTimeFormatList = value ?? FORMATTED_TIME_FORMAT_LIST;

  factory ZonedDateTime.fromJson(String json) => json != null ? ZonedDateTime(date: json) : null;
  factory ZonedDateTime.fromJsonMap(Map<String, dynamic> json) => json != null ? ZonedDateTime(date: json['date'] as String) : null;
}